--- hc.lua plugin for STI
-- @module hc.lua
-- @author Martin Braun (modiX)
-- @copyright 2018
-- @license MIT/X11

local lg = require((...):gsub('plugins.hc', 'graphics'))

return {
	hc_LICENSE        = "MIT/X11",
	hc_URL            = "https://github.com/karai17/Simple-Tiled-Implementation",
	hc_DESCRIPTION    = "HadronCollider hooks for STI.",

	--- Adds each collidable tile to the hadron collider.
	-- @param world The tiled map object.
	-- @return collidables table containing the handles to the objects in the Bump world.
	hc_init = function(map, collider)
		local shapes = {}
		
		for _, layer in ipairs(map.layers) do
			-- Entire layer
			if layer.properties.collidable == true then
				if layer.type == "tilelayer" then
					for y, tiles in ipairs(layer.data) do
						for x, tile in pairs(tiles) do
							if tile.objectGroup then
								for _, object in ipairs(tile.objectGroup.objects) do
									if object.properties.collidable then
										local t = {
											x = ((x-1) * map.tilewidth  + tile.offset.x + map.offsetx) + object.x,
											y = ((y-1) * map.tileheight + tile.offset.y + map.offsety) + object.y,
											w = object.width,
											h = object.height
										}
										local shape
										if object.shape == "rectangle" then
											shape = collider:rectangle(t.x, t.y, t.w, t.h)
										elseif object.shape == "polyline" then
											local args = {}
											for i, p in ipairs(object.polyline) do
												if i < #object.polyline then
													--if #args < 2 or args[#args-1] ~= p.x or args[#args] ~= p.y then
														table.insert(args, t.x + p.x)
														table.insert(args, t.y + p.y)
													--end
												end
											end
											shape = collider:polygon(unpack(args))
										end
										if shape then
											shape.layer = layer
											shape.properties = object.properties
											table.insert(shapes, shape)
										end
									end
								end
							end
						end
					end
				end
			end
		end
		map.hc_shapes = shapes
	end,

	--- Remove layer
	-- @param index to layer to be removed
	-- @param world bump world the holds the tiles
	-- @param tx Translate on X
-- @param ty Translate on Y
-- @param sx Scale on X
-- @param sy Scale on Y
	hc_removeLayer = function(map, index, collider)
		local layer = assert(map.layers[index], "Layer not found: " .. index)
		local shapes = map.hc_shapes

		-- Remove collision objects
		for i = #shapes, 1, -1 do
			local obj = shapes[i]
			if obj.layer == layer and (layer.properties.collidable == true	or obj.properties.collidable == true) then
				collider:remove(obj)
				table.remove(shapes, i)
			end
		end
	end,

	--- Draw bump collisions world.
	-- @param world bump world holding the tiles geometry
	-- @param tx Translate on X
	-- @param ty Translate on Y
	-- @param sx Scale on X
	-- @param sy Scale on Y
	hc_draw = function(map, tx, ty, sx, sy, fillColor, outlineColor)
		lg.push()
		lg.scale(sx or 1, sy or sx or 1)
		--lg.translate(tx, ty)

		for _, shape in ipairs(map.hc_shapes) do
			love.graphics.setColor(fillColor)
			shape:draw("fill")
			love.graphics.setColor(outlineColor)
			shape:draw("line")
			love.graphics.setColor({ 255, 255, 255, 255 })
		end

		lg.pop()
	end
}
